/**
 * 
 *    Copyright 2013 EOSSOnline Limited
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  
**/
package com.eossonline.edrive.commands;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.chemistry.opencmis.commons.impl.json.JSONObject;
import org.apache.chemistry.opencmis.commons.impl.json.parser.JSONParser;
import org.apache.commons.lang.WordUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.eossonline.edrive.service.EDrive;
import com.eossonline.esync.ESync;
import com.eossonline.utils.json.JSONWriter;

public class CommandFactory {
	
	private static final Logger logger = LoggerFactory.getLogger(CommandFactory.class);

	private Pattern commandSplitter = Pattern.compile("^\\s*([\\w|-]+)\\s*(.*)$",Pattern.CASE_INSENSITIVE);
	
	private static Pattern methods = Pattern.compile("^(add|cancel-checkout|checkin|checkout|disable|enable|get|remove|save|set|test|toggle|update|shutdown)$");
	
	/*
	 * Methods that do not require commands do not have entries here.
	 * For instance, remove only requires the parameters object containing a name element
	 */
	private static Pattern addCommands = Pattern.compile("^(mime-mapping|synchroniser|ignore-pattern)$");			
	private static Pattern getCommands = Pattern.compile("^(document|defaults|configuration|thumb-nails|version-history)$");
	private static Pattern removeCommands = Pattern.compile("^(mime-mapping|synchroniser|ignore-pattern)$");
	private static Pattern setCommands = Pattern.compile("^(default-type|proxy|sync-seconds|sync-direction|toggle|sync-path|enable-unfiling|repository-delete)$");
	private static Pattern testCommands = Pattern.compile("^(connection)$");
	private static Pattern updateCommands = Pattern.compile("^(configuration)$");
	
	private static Map<String, Pattern> commands = new HashMap<String, Pattern>();
	static {
		commands.put("add", addCommands);
		commands.put("get", getCommands);
		commands.put("set", setCommands);
		commands.put("remove", removeCommands);
		commands.put("test", testCommands);
		commands.put("update", updateCommands);		
	}
	
	
	private static CommandFactory instance = new CommandFactory();
	
	public static final String RESULT = "result";
	public static final String NAME = "name";
	
	public static CommandFactory getInstance() {
		return CommandFactory.instance;
	}
	
	
	public JSONObject processRequest(JSONParser parser, String request) {

		JSONObject params = null;

		String method = null;
		String requestParams = null;

		JSONObject result = null;

		// Tokenise the command string into the command and any parameters
		Matcher matcher = commandSplitter.matcher(request);
		if (matcher.matches()) {
			method = matcher.group(1).toLowerCase();
			Matcher c = methods.matcher(method);
			if (!c.matches()) {
				// not a method we understand
				// Don't know what this is but we won't handle it
				params = new JSONObject();
				params.put("result", "Unknown method [" + method + "].");
				return generateErrorResponse(params);
			}
			
			requestParams = matcher.group(2);
		}

		String command = null;
		Pattern cmd = commands.get(method);
		if(cmd == null) {
			command ="";
			// No command required so the remainder should be the request params
			requestParams = matcher.group(2);
		} else {
			String validMethods = cmd.toString().replace("^", "").replace("$", "");
			matcher = commandSplitter.matcher(requestParams);
			if(matcher.matches()) {
				command = matcher.group(1).toLowerCase();
				Matcher c = cmd.matcher(command);
				if(!c.matches()) {
					// not a method we understand
					// Don't know what this is but we won't handle it
					params = new JSONObject();
					params.put("result", "Unknown command [" + command + "] for method [" + method + "]. Valid commands are " + validMethods + ".");
					return generateErrorResponse(params);
				}
				
				requestParams = matcher.group(2);
			} else {
				JSONObject obj = new JSONObject();
				obj.put("result", "You must provide a valid command for method [" + method + "]. Valid commands are " + validMethods + ".");
				return generateErrorResponse(obj);
			}
		}

		if((requestParams == null) || (requestParams.length() == 0)) {
			requestParams = "{}";
		}

		// We now have a valid method / command combination
		// and requestParams should be a valid JSONObbject
		try {
			params = (JSONObject) parser.parse(requestParams);
		} catch (Exception e) {
			return generateExceptionResponse(new CommandException( requestParams + " is not a valid JSON Object."));
		}

		String factoryClassName = WordUtils.capitalizeFully(method + "-" + command, new char [] {'-'}).replaceAll("-", "");
		Class<?> c = null;
		Object o = null;
		try {
			try {
				c = Class.forName(getClass().getPackage().getName() + "." + factoryClassName);
				o = c.newInstance();
				if(!(o instanceof Command)) {
					throw new ClassNotFoundException();
				}
			} catch(ClassNotFoundException cnfe) {
				return generateExceptionResponse(new CommandException(
						"No Command Class can be found for [" + factoryClassName + "]." +
								" It' full name MUST be [" + getClass().getPackage().getName() + "." + factoryClassName + "]" +
								" and it MUST implement the Interface [" + Command.class.getCanonicalName() + "]"
						));
			} catch(Throwable e) {
				return generateExceptionResponse(e);
			} 

			Method m = c.getDeclaredMethod("execute", EDrive.class, JSONObject.class);
			result = (JSONObject)m.invoke(o, EDrive.getEDrive(), params);
			
		}catch(Exception e) {
			return generateExceptionResponse(e);
		} finally {
			// Help out the garbage collector. Should not be necessary but it won't hurt :-)
			o = null;
		}
		
		return generateResponse(result);
	}
	

	public static JSONObject generateResponse(JSONObject obj) {
		JSONObject result = new JSONObject();

		String status = (obj.get("status") == null) ? "OK" : (String) obj.get("status");
		obj.remove("status");
		String message = (String)obj.get("message");
		if(message != null) {
			result.put("message", message);
			obj.remove("message");
		}
		result.put("status", status);
		result.put("result", obj);

		if(logger.isDebugEnabled()) {
			JSONWriter writer = new JSONWriter(); // this is the new writer that adds  indentation.
			try {
				result.writeJSONString(writer);
				logger.debug("OK : \n{}", writer.toString());
			}catch(IOException ioe) {
				logger.debug("OK : \n{}", result);				
			}
		}
		return result;
	}

	public static JSONObject generateErrorResponse(JSONObject obj) {
		JSONObject result = new JSONObject();
		result.put("status", "ERROR");
		result.put("result", obj.get("result"));
		if(obj.get("message") != null) {
			result.put("message", (String)obj.get("message"));
		}
		// Remove any message from the object as we don't actually want it shown as part of the detail
		if(logger.isDebugEnabled()) {
			JSONWriter writer = new JSONWriter(); // this is the new writer that adds  indentation.
			try {
				result.writeJSONString(writer);
				logger.debug("ERROR : \n{}", writer.toString());
			}catch(IOException ioe) {
				logger.debug("ERROR : \n{}", result);				
			}
		}
		return result;
	}

	public static JSONObject generateExceptionResponse(Throwable t) {
		JSONObject result = new JSONObject();
		result.put("status", "EXCEPTION");
		
		if(t instanceof InvocationTargetException) {
			t = ((InvocationTargetException)t).getTargetException();
		}

		String message = t.getLocalizedMessage();
		if(message == null) {
			message = t.toString();
		}
		StringBuilder sb = new StringBuilder(message);
		/*
		for(StackTraceElement ste : t.getStackTrace()) {
			sb.append("\n").append(ste);
		}
		*/
		result.put("result", sb.toString());
		if(logger.isDebugEnabled()) {
			JSONWriter writer = new JSONWriter(); // this is the new writer that adds  indentation.
			try {
				result.writeJSONString(writer);
				logger.debug("EXCEPTION : \n{}", writer.toString());
			}catch(IOException ioe) {
				logger.debug("EXCEPTION : \n{}", result);				
			}
		}
		return result;
	}

	public static ESync getSynchroniser(String name) throws CommandException {
		ESync eSync = EDrive.getEDrive().getSynchroniser(name);
		if(eSync == null) {
			throw new CommandException("No Synchroniser registered with name [" + name + "]");
		}
		return eSync;
	}

}


/*
 * A couple of utility exceptions
 */
class UnsupportedCommandException extends RuntimeException {
	
	private static final long serialVersionUID = -4345683141623793699L;

	UnsupportedCommandException(String message) {
		super(message);
	}
}

class MissingPropertyException extends RuntimeException {
	
	private static final long serialVersionUID = -4451949158053868887L;

	MissingPropertyException(String message) {
		super(message);
	}
}

class CommandException extends Exception {
	
	private static final long serialVersionUID = 1L;

	CommandException(String message) {
		super(message);
	}
	
	CommandException(String message, Throwable t) {
		super(message, t);
	}
}
