/**
 * 
 *    Copyright 2013 EOSSOnline Limited
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  
**/
package com.eossonline.edrive.commands;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.chemistry.opencmis.client.api.Repository;
import org.apache.chemistry.opencmis.client.api.SessionFactory;
import org.apache.chemistry.opencmis.client.runtime.SessionFactoryImpl;
import org.apache.chemistry.opencmis.commons.SessionParameter;
import org.apache.chemistry.opencmis.commons.enums.VersioningState;
import org.apache.chemistry.opencmis.commons.impl.json.JSONArray;
import org.apache.chemistry.opencmis.commons.impl.json.JSONObject;
import org.apache.commons.codec.binary.Base64;

import com.eossonline.edrive.service.EDrive;
import com.eossonline.esync.ESync;
import com.eossonline.esync.SyncDirectionEnum;



/*
 * NOTE: All Command classes should instantiate a static instance of the class that is returned
 * by the CommandFactory. This ensures that we are not creating new unnecessary class instances over time
 * that would constitute an effective memory leak. This way we don't have to worry about deleting them later.
 */


class GetDocument implements Command {

	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {

		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		
		if(obj.get("id") == null) {
			throw new CommandException("You must define a valid document id element.");			
		}
		String name = (String)obj.get(CommandFactory.NAME);
		ESync eSync = CommandFactory.getSynchroniser(name);
		
		return eDrive.getDocument(eSync, (String)obj.get("id"));
	}
}

class GetThumbNails implements Command {

	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {

		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element.");			
		}
		Path path = Paths.get((String)obj.get("path"));
		String name = (String)obj.get(CommandFactory.NAME);
		ESync eSync = CommandFactory.getSynchroniser(name);
		
		return eDrive.getThumbNails(eSync, path);
	}
}

class GetVersionHistory implements Command {

	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {

		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element.");			
		}
		Path path = Paths.get((String)obj.get("path"));
		String name = (String)obj.get(CommandFactory.NAME);
		ESync eSync = CommandFactory.getSynchroniser(name);
		
		return eDrive.getVersionHistory(eSync, path);
	}
}

class Shutdown implements Command {

	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		eDrive.shutdown();
		return new JSONObject();
	}
}

/**
* Set proxy user name and password if using a proxy.
* 
* Some Repositories are configured behind proxies. This will configure all or a single named Synchroniser 
* with the proxy user name and password
* 
* @author smcardle
*
*/
class SetProxy implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		
		// We must also have a valid path for this Synchroniser
		if(obj.get("proxyUserName") == null) {
			throw new CommandException("You must define a valid proxyUserName element.");
		}
		String proxyUserName = (String)obj.get("proxyUserName");
		
		// We must also have a valid path for this Synchroniser
		if(obj.get("proxyPassword") == null) {
			throw new CommandException("You must define a valid proxyPassword element.");
		}
		String proxyPassword = (String)obj.get("proxyPassword");
		
		if(name == null) {
			// ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				eSync.setProxyDetails(proxyUserName, proxyPassword);
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			eSync.setProxyDetails(proxyUserName, proxyPassword);
			return eSync.getJSONSyncConfiguration();
		}
	}
}

/**
* Remove a mime mapping that allows a specific document type to be used if the Document being created
* in the Repository is of this any of the defined mime types.
* 
* @author smcardle
*
*/
class RemoveMimeMapping implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				if(obj.get("mimeMappings") != null) {
					JSONArray ja = (JSONArray)obj.get("mimeMappings");
					for(Object o : ja) {
						removeMimeMappings(eSync, (JSONObject)o);
					}
				} else {
					removeMimeMappings(eSync, obj);
				}
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			// ALL the Synchronisers
			if(obj.get("mimeMappings") != null) {
				JSONArray ja = (JSONArray)obj.get("mimeMappings");
				for(Object o : ja) {
					removeMimeMappings(eSync, (JSONObject)o);
				}
			} else {
				removeMimeMappings(eSync, obj);
			}
			return eSync.getJSONSyncConfiguration();
		}
	}
	
	private void removeMimeMappings(ESync eSync, JSONObject obj) throws CommandException {
		if(obj.get("documentType") == null) {
			throw new CommandException("You must define a valid documentType element.");
		}
		String documentType = (String)obj.get("documentType");
		if(obj.get("mimeTypes") == null) {
			throw new CommandException("You must define a valid mimeTypes element.");
		}
		String mimeTypes = (String)obj.get("mimeTypes");
		
		eSync.getSyncConfiguration().removeMimeMappings(documentType, mimeTypes);
	}
}



/**
* Add a mime mapping that allows a specific document type to be used if the Document being created
* in the Repository is of this any of the defined mime types.
* 
* @author smcardle
*
*/
class AddMimeMapping implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				if(obj.get("mimeMappings") != null) {
					JSONArray ja = (JSONArray)obj.get("mimeMappings");
					for(Object o : ja) {
						addMimeMappings(eSync, (JSONObject)o);
					}
				} else {
					addMimeMappings(eSync, obj);
				}
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			// ALL the Synchronisers
			if(obj.get("mimeMappings") != null) {
				JSONArray ja = (JSONArray)obj.get("mimeMappings");
				for(Object o : ja) {
					addMimeMappings(eSync, (JSONObject)o);
				}
			} else {
				addMimeMappings(eSync, obj);
			}
			return eSync.getJSONSyncConfiguration();
		}
	}
	
	private void addMimeMappings(ESync eSync, JSONObject obj) throws CommandException {
		if(obj.get("documentType") == null) {
			throw new CommandException("You must define a valid documentType element.");
		}
		String documentType = (String)obj.get("documentType");
		if(obj.get("mimeTypes") == null) {
			throw new CommandException("You must define a valid mimeTypes element.");
		}
		
		eSync.getSyncConfiguration().addMimeMappings(documentType, (String)obj.get("mimeTypes"));
	}
}

class Checkin implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);

		// We must also have a valid path for this Synchroniser
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element. This can be an absolute path or a relative path from the fileSystemRoot element.");
		}		
		Path dir = Paths.get((String)obj.get("path"));
		if(!dir.isAbsolute()) {
			dir = eSync.getFileSystemSyncPath().resolve(dir);
		} else {
			// Could be either an absolute File System path or Repository path
			if(dir.startsWith(eSync.getRepositorySyncPath())) {
				dir = eSync.getFileSystemSyncPath().resolve(eSync.getRepositorySyncPath().relativize(dir));
			}
		}
		
		VersioningState versioningState = VersioningState.MINOR;
		if(obj.get("version") != null && "MAJOR".equals((String)obj.get("version"))) {
			versioningState = VersioningState.MAJOR;
		}
		String comment = "eDrive - Version checked in by [" + eDrive.getConfiguration().get("repositoryUserName")+ "]";
		if(obj.get("comment") != null) {
			comment = (String)obj.get("comment");
		}
		
		if(eSync.isCheckedOut(dir)) {
			eSync.checkin(dir, versioningState, comment);
		} else {
			throw new CommandException("Document [" + ESync.getPathAsString(dir) + "] is not Checked Out.");
		}
		
		return eSync.getJSONSyncConfiguration();
	}
}

class CancelCheckout implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);

		// We must also have a valid path for this Synchroniser
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element. This can be an absolute path or a relative path from the fileSystemRoot element.");
		}		
		Path dir = Paths.get((String)obj.get("path"));
		
		if(!dir.isAbsolute()) {
			dir = eSync.getFileSystemSyncPath().resolve(dir);
		} else {
			// Could be either an absolute File System path or Repository path
			if(dir.startsWith(eSync.getRepositorySyncPath())) {
				dir = eSync.getFileSystemSyncPath().resolve(eSync.getRepositorySyncPath().relativize(dir));
			}
		}
		
		if(eSync.isCheckedOut(dir)) {
			eSync.cancelCheckout(dir);
		} else {
			throw new CommandException("Document [" + ESync.getPathAsString(dir) + "] is not Checked Out.");
		}
		
		return eSync.getJSONSyncConfiguration();
	}
}

class Checkout implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);

		// We must also have a valid path for this Synchroniser
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element. This can be an absolute path or a relative path from the fileSystemRoot element.");
		}		
		Path dir = Paths.get((String)obj.get("path"));
		
		if(!dir.isAbsolute()) {
			dir = eSync.getFileSystemSyncPath().resolve(dir);
		} else {
			// Could be either an absolute File System path or Repository path
			if(dir.startsWith(eSync.getRepositorySyncPath())) {
				dir = eSync.getFileSystemSyncPath().resolve(eSync.getRepositorySyncPath().relativize(dir));
			}
		}
		
		if(!eSync.isCheckedOut(dir)) {
			eSync.checkout(dir);
		} else {
			throw new CommandException("Document [" + ESync.getPathAsString(dir) + "] is already Checked Out.");
		}
		
		return eSync.getJSONSyncConfiguration();
	}
}


/**
* Toggle a path tree from a single path.
* 
* If the path does not exist this has no effect
* 
* If the path is currently set to syncing true it will be set to false and vis versa
* It will also set EVERY child path from this path down the monitored hierarchy to the
* same value as this path was set irrespective of each paths previous setting 
* 
* @author smcardle
*
*/
class Toggle implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);

		// We must also have a valid path for this Synchroniser
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element. This can be an absolute path or a relative path from the fileSystemRoot element.");
		}		
		Path dir = Paths.get((String)obj.get("path"));
		eSync.toggleSync(dir);
		
		return eSync.getJSONSyncConfiguration();
	}
}

/**
* Set whether unfiling is enabled or not.
* 
* Some Repositories do not support unfiling so trying to enable it be ignored
* 
* @author smcardle
*
*/
class SetEnableUnfiling implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		// We must also have a valid path for this Synchroniser
		if(obj.get("unfilingEnabled") == null) {
			throw new CommandException("You must define a valid unfilingEnabled element.  This is a boolean true or false.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);
		Boolean unfiling = (Boolean)obj.get("unfilingEnabled");
		
		// We now have all the parts we need
		eSync.enableUnfiling(unfiling);
		
		return eSync.getJSONSyncConfiguration();
	}
}

/**
* Set whether to enable deleting of Documents from the Repository.
* 
* Sometimes we do not want to allow deletes in the Repository. If this is set to false
* the likely result will be that the Document in the Repository will be re-created on
* the File System.
* 
* Obviously, the Repository is able to move the Document, possibly using rules, and
* if the sync direction is either FROM_REPOSITORY, TO_REPOSITORY_DELETE_SOURCE or BI_DIRECTIONAL
* the file corresponding to the Document in the Repository will be deleted
* 
* @author smcardle
*
*/
class SetRepositoryDelete implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		// We must also have a valid path for this Synchroniser
		if(obj.get("repositoryDeleteEnabled") == null) {
			throw new CommandException("You must define a valid repositoryDeleteEnabled element.  This is a boolean true or false.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);
		Boolean deleteEnabled = (Boolean)obj.get("repositoryDeleteEnabled");
		
		// We now have all the parts we need
		eSync.enableRepositoryDelete(deleteEnabled);
		
		return eSync.getJSONSyncConfiguration();
	}
}

/**
* Set a specific path to enable or disable Synchronising.
* 
* If the path does not exist this has no effect
* 
* @author smcardle
*
*/
class SetSyncPath implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a valid name element
		if(obj.get("name") == null) {
			throw new CommandException("You must define a valid name element.");
		}
		// We must also have a valid path for this Synchroniser
		if(obj.get("path") == null) {
			throw new CommandException("You must define a valid path element. This can be an absolute path or a relative path from the fileSystemRoot element.");
		}
		if(obj.get("syncing") == null) {
			throw new CommandException("You must define a valid syncing element. This is a boolean true or false.");
		}
		String name = (String)obj.get(CommandFactory.NAME);
		// If the named Synchroniser does not exist it will throw an exception
		ESync eSync = CommandFactory.getSynchroniser(name);
		Path dir = Paths.get((String)obj.get("path"));
		Boolean syncing = (Boolean)obj.get("syncing");
		
		// We now have all the parts we need
		eSync.setSyncPath(dir, syncing);
		
		return eSync.getJSONSyncConfiguration();
	}
}

/**
* Set the direction the Synchroniser is configured
* 
* The syncDirection element needs to be one of TO_REPOSITORY, FROM_REPOSITORY, TO_RESPOSITORY_DELETE_SOURCE,
* BI_DIRECTION or NO_SYNC
* 
* @author smcardle
*
*/
class SetSyncDirection implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		// Must have a syncDirection element
		if(obj.get("syncDirection") == null) {
			throw new CommandException("You must define a valid syncDirection element. This can be either TO_REPOSITORY, TO_REPOSITORY_DELETE_SOURCE, FROM_REPOSITORY, BI_DIRECTIONAL or NO_SYNC");
		}
		String sd = (String)obj.get("syncDirection");
		SyncDirectionEnum syncDirection = null;
		try {
			syncDirection = SyncDirectionEnum.valueOf(sd);
		}catch(Exception e) {
			throw new CommandException("The syncDirection value [" + sd + "] is not valid. This should be either TO_REPOSITORY, TO_REPOSITORY_DELETE_SOURCE, FROM_REPOSITORY, BI_DIRECTIONAL or NO_SYNC");
		}
				
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// Enable ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				eSync.setSyncDirection(syncDirection);
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			eSync.setSyncDirection(syncDirection);
			return eSync.getJSONSyncConfiguration();
		}
	}
}



/**
 * Set either the Inbound and Outbound sync seconds
 * Either the element inboundSyncSeconds or outboundSyncSeconds needs to be defined with a value.
 * This method will ignore values less than 1, making 1 the minimum number of seconds
 * 
 * @author smcardle
 *
 */
class SetSyncSeconds implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		long inboundSeconds = -1L;
		long outboundSeconds = -1L;
		// Must have a sync-seconds element
		if(obj.get("outboundSyncSeconds") != null) {
			outboundSeconds = ((Number)obj.get("outboundSyncSeconds")).longValue();
		}
		if(obj.get("inboundSyncSeconds") != null) {
			inboundSeconds = ((Number)obj.get("inboundSyncSeconds")).longValue();
		}
		
		if(inboundSeconds == -1 && outboundSeconds == -1) {
			throw new CommandException("You must specify a property inboundSyncSeconds or outboundSyncSeconds or both");
		}
				
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				if(inboundSeconds > 0)
					eSync.getSyncConfiguration().setInboundSyncSeconds(inboundSeconds);
				if (outboundSeconds > 0)
					eSync.getSyncConfiguration().setOutboundSyncSeconds(outboundSeconds);
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			if(inboundSeconds > 0)
				eSync.getSyncConfiguration().setInboundSyncSeconds(inboundSeconds);
			if (outboundSeconds > 0)
				eSync.getSyncConfiguration().setOutboundSyncSeconds(outboundSeconds);
			return eSync.getJSONSyncConfiguration();
		}
	}
}


/**
 * Filter patterns have to be sent and received as Base64 encoded. This is because the special
 * characters i.e. wild cards, in regular expressions are not able to be placed directly into
 * JSON Strings
 * 
 * @author smcardle
 *
 */
class RemoveIgnorePattern implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		JSONArray ignorePatterns = new JSONArray();
		// obj must have an Array of ignorePatterns OR a single ignorePattern
		try {
		if(obj.get("ignorePatterns") == null) {
			if(obj.get("base64") == null) {
				throw new CommandException("You muse define either an Array of ignorePatterns or a single base64 element.");
			}
			String base64 = (String)obj.get("base64");
			JSONObject o = new JSONObject();
			o.put("base64", base64);
			ignorePatterns.add(o);
		} else {
			ignorePatterns = (JSONArray)obj.get("ignorePatterns");
		}
		}catch(Exception e) {
			throw new CommandException(e.getLocalizedMessage());
		}
		
		// We now have an Array of ignorePatterns

		Base64 base64 = new Base64();


		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
	    	for(Object ip : ignorePatterns) {
	    		JSONObject jo = (JSONObject)ip;
	    		Pattern p = Pattern.compile(new String(base64.decode((String)jo.get("base64"))));
				// Adds the filter pattern to ALL the Synchronisers
				for(ESync eSync : eDrive.getSynchronisers()) {
					eSync.getSyncConfiguration().removeIgnorePattern(p);
				}
	    	}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
	    	for(Object ip : ignorePatterns) {
	    		JSONObject jo = (JSONObject)ip;
	    		Pattern p = Pattern.compile(new String(base64.decode((String)jo.get("base64"))));
	    		eSync.getSyncConfiguration().removeIgnorePattern(p);
	    	}		
			return eSync.getJSONSyncConfiguration();
		}
	}
}

/**
 * Filter patterns have to be sent and received as Base64 encoded. This is because the special
 * characters i.e. wild cards, in regular expressions are not able to be placed directly into
 * JSON Strings
 * 
 * @author smcardle
 *
 */
class AddIgnorePattern implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		JSONArray ignorePatterns = new JSONArray();
		// obj must have an Array of ignorePatterns OR a single ignorePattern
		try {
		if(obj.get("ignorePatterns") == null) {
			if(obj.get("base64") == null) {
				throw new CommandException("You muse define either an Array of ignorePatterns or a single base64 element.");
			}
			String base64 = (String)obj.get("base64");
			JSONObject o = new JSONObject();
			o.put("base64", base64);
			ignorePatterns.add(o);
		} else {
			ignorePatterns = (JSONArray)obj.get("ignorePatterns");
		}
		}catch(Exception e) {
			throw new CommandException(e.getLocalizedMessage());
		}
		
		// We now have an Array of ignorePatterns

		Base64 base64 = new Base64();


		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
	    	for(Object ip : ignorePatterns) {
	    		JSONObject jo = (JSONObject)ip;
	    		Pattern p = Pattern.compile(new String(base64.decode((String)jo.get("base64"))));
				// Adds the filter pattern to ALL the Synchronisers
				for(ESync eSync : eDrive.getSynchronisers()) {
					eSync.getSyncConfiguration().addIgnorePattern(p);
				}
	    	}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
	    	for(Object ip : ignorePatterns) {
	    		JSONObject jo = (JSONObject)ip;
	    		Pattern p = Pattern.compile(new String(base64.decode((String)jo.get("base64"))));
	    		eSync.getSyncConfiguration().addIgnorePattern(p);
	    	}		
			return eSync.getJSONSyncConfiguration();
		}
	}
}

class TestConnection implements Command {
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {

		StringBuilder sb = new StringBuilder();
		String url = null;

    	Map<String, String> params = new HashMap<String,String>();

		String space = "";
		if(obj.get("repositoryUrl") == null) {
			sb.append("repositoryUrl").append(space);
			space = " ";
		} else {
			url = (String)obj.get("repositoryUrl");
		}
		if(obj.get("repositoryUserName") == null) {
			sb.append("repositoryUserName").append(space);
			space = " ";
		} else {
			params.put(SessionParameter.USER, (String)obj.get("repositoryUserName"));
		}
		if(obj.get("repositoryPassword") == null) {
			sb.append("repositoryPassword").append(space);
			space = " ";
		} else {
			params.put(SessionParameter.PASSWORD, (String)obj.get("repositoryPassword"));
		}		
		if(sb.length() != 0) {
			throw new CommandException("Unable to test the connection. Missing required elements [" + sb.toString() + "]");
		}
		
		// Optional parameters
		
		if(obj.get("repositoryId") != null) {
			params.put(SessionParameter.REPOSITORY_ID, (String)obj.get("repositoryId"));
		}
		
		Boolean getFolders = false;
		if(obj.get("getFolders") != null) {
			getFolders = (Boolean)obj.get("getFolders");
		}
		
		Path startFolder = null;
		if(obj.get("startFolder") != null) {
			startFolder = Paths.get((String)obj.get("startFolder"));
		}
		if(obj.get("proxyUserName") != null) {
			params.put(SessionParameter.PROXY_USER,(String)obj.get("proxyUserName"));
		}
		if(obj.get("proxyPassword") != null) {
			params.put(SessionParameter.PROXY_PASSWORD,(String)obj.get("proxyPassword"));
		}
		
		// NOTE: Only one start folder can be defined. This means that if the CMS supports multiple Repositories it's likely that
		// this path will match only ONE of them. In this case, ALL other repositories will list folders from their ROOT.
		// To mitigate this it's best to provide the repositoryId as well so that ONLY that repository is searched.
		
		/*
		 * A usage pattern could be to get ALL the repositories only, then after selecting one get the Folders from startFolder or from the RootFolder
		 */
		
		Map<String, String> parameters = ESync.getRepositoryWebServiceParameters(url);
		parameters.putAll(params);

		Boolean acceptAllCerts = false;
		if(obj.get("acceptAllCerts") != null) {
			acceptAllCerts = (Boolean)obj.get("acceptAllCerts");
		}
		
		if(acceptAllCerts) {
			
		}
		
		
		// Got all the required bits and maybe an ID, so lets try and see if we can connect to the repository
    	SessionFactory sessionFactory = SessionFactoryImpl.newInstance();
    	//If no repositoryId is set then a full list of repositories will be returned, else, just the one we asked for
    	List<Repository> repos = null;
    	try {
    		repos = sessionFactory.getRepositories(parameters);
    	}catch(Exception e) {
    		throw new CommandException("Repository URL [" + url + "]. " +  e.getMessage());
    	}
    	JSONObject result = new JSONObject();
    	JSONArray ja = new JSONArray();
    	for(Repository repo : repos) {
    		JSONObject o = new JSONObject();
    		o.put("id", repo.getId());
    		o.put("vendor", repo.getVendorName());
    		o.put("description", repo.getDescription());
    		o.put("productName", repo.getProductName());
    		o.put("productVersion", repo.getProductVersion());
    		if(getFolders) {
    			// This can take some time and is not always required
	    		JSONArray af = eDrive.getAvailableFolders(repo, startFolder);
	    		o.put("folders", af);
    		}
    		ja.add(o);
    	}
    	result.put("repositories", ja);
		
		return result;
	}	
}

/**
 * Disable a specified Synchoniser or ALL of them
 * @author smcardle
 *
 */
class Disable implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// Disable ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				eSync.disable();
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			eSync.disable();
			return eSync.getJSONSyncConfiguration();
		}
	}
}

/**
 * Enable a specified Synchoniser or ALL of them
 * @author smcardle
 *
 */
class Enable implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// Enable ALL the Synchronisers
			for(ESync eSync : eDrive.getSynchronisers()) {
				eSync.enable();
			}
			return eDrive.getConfiguration();
		} else {
			ESync eSync = CommandFactory.getSynchroniser(name);
			eSync.enable();
			return eSync.getJSONSyncConfiguration();
		}
	}
}

/**
 * Save the current Configuration.
 * @author smcardle
 *
 */
class Save implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		eDrive.saveSyncConfig();
		return eDrive.getConfiguration();
	}
}

/**
 * Remove a Synchroniser Configuration. This will shut it down and remove it
 * @author smcardle
 *
 */
class RemoveSynchroniser implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			throw new CommandException("You MUST have a name attribute to identify this Configuration by.");
		}
		if(eDrive.getSynchroniser(name) == null) {
			throw new CommandException("The Sychroniser identified by name [" + name + "] does not exist.");
		}
		eDrive.removeSynchroniser(name);
		
		// return the new current configuration
		return eDrive.getConfiguration();
	}
}

/**
 * Set the default types to use for Folders and Documents when creating entries in the Repository
 * @author smcardle
 *
 */
class SetDefaultType implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			throw new CommandException("You MUST have a name attribute to identify the Configuration by.");
		}
		ESync eSync = CommandFactory.getSynchroniser(name);
		
		if(obj.get("defaultDocumentType") != null)
			eSync.getSyncConfiguration().setDocumentType((String)obj.get("defaultDocumentType"));
		if(obj.get("defaultFolderType") != null)
			eSync.getSyncConfiguration().setFolderType((String)obj.get("defaultFolderType"));		
		
		// return the new current configuration
		return eSync.getJSONSyncConfiguration();
	}
}

/**
 * Add a new Synchroniser Configuration
 * @author smcardle
 *
 */
class AddSynchroniser implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			throw new CommandException("You MUST have a name attribute to identify this Configuration by.");
		}
		ESync eSync = eDrive.addSynchroniser(obj);
		return eSync.getJSONSyncConfiguration();
	}
}

/**
 * Update any part of an existing configuration.
 * 
 * @author smcardle
 *
 */
class UpdateConfiguration implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// This means update these values in EVERY registered Synchroniser
			return eDrive.updateConfiguration(obj);
		}
		ESync eSync = CommandFactory.getSynchroniser(name);
		return eSync.updateConfiguration(obj);
	}
}

class SetConfiguration implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {
		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// This means update these values in EVERY registered Synchroniser
			return eDrive.setConfiguration(obj);
		}
		ESync eSync = CommandFactory.getSynchroniser(name);
		return eSync.setConfiguration(obj);
	}	
}


class GetConfiguration implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj) throws CommandException {

		String name = (String)obj.get(CommandFactory.NAME);
		if(name == null) {
			// Get ALL configurations
			return eDrive.getConfiguration();
		} 
		ESync eSync = CommandFactory.getSynchroniser(name);
		return eSync.getJSONSyncConfiguration();
	}
}

class GetDefaults implements Command {
	
	public JSONObject execute(EDrive eDrive, JSONObject obj)  throws CommandException {
		return eDrive.getConfigurationDefaults();
	}
}

