/**
 * 
 *    Copyright 2013 EOSSOnline Limited
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  
**/
package com.eossonline.edrive.events;

import java.nio.file.Path;

import org.apache.chemistry.opencmis.client.api.Document;
import org.apache.chemistry.opencmis.client.api.CmisObject;


import com.eossonline.edrive.events.EDriveEvent;
import com.eossonline.edrive.service.EDrive;
import com.eossonline.esync.ESync;
import com.eossonline.esync.filesystem.OutboundProcessor;
import com.eossonline.esync.repository.InboundProcessor;

privileged public aspect EDriveEventAspect {
	
	pointcut syncInbound(InboundProcessor source) : 
		execution(private void InboundProcessor.syncRepository())
		&& target(source);
	
	pointcut syncOutbound(OutboundProcessor source) : 
		execution(void OutboundProcessor.syncFileSystem())
		&& target(source);
	
	pointcut createDocumentOutbound(ESync source, Path path) :
		execution(void ESync.createInRepository(Path))
		&& target(source)
		&& args(path);

	pointcut createDocumentInbound(ESync source, Document document) :
		execution(void ESync.createInFileSystem(Document))
		&& !cflow(execution(void ESync.updateInFileSystem(*)))
		&& target(source)
		&& args(document);

	pointcut updateDocumentOutbound(ESync source, Path path) :
		execution(void ESync.updateInRepository(Path))
		&& target(source)
		&& args(path);
	
	pointcut updateDocumentInbound(ESync source, Document document) :
		execution(void ESync.updateInFileSystem(CmisObject))
		&& target(source)
		&& args(document);
	
	pointcut deleteDocumentOutbound(ESync source, Path path) :
		execution(void ESync.deleteFromRepository(Path))
		&& target(source)
		&& args(path);
	
	pointcut deleteDocumentInbound(ESync source, Path path) :
		execution(void ESync.deleteFromFileSystem(Path))
		&& target(source)
		&& args(path);
	
	pointcut addFolder(OutboundProcessor source, Path path, Boolean syncing) :
		execution(Boolean addFolder(Path, Boolean))
		&& this(source)
		&& args(path, syncing);
	
	pointcut removeFolder(OutboundProcessor source, Path path, Boolean syncing) :
		execution(Boolean removeFolder(Path, Boolean))
		&& this(source)
		&& args(path, syncing);

	pointcut checkoutDocument(ESync source, Path path) :
		execution(Document ESync.checkout(Path))
		&& target(source)
		&& args(path);
	
	pointcut checkinDocument(ESync source, Path path) :
		execution(Document ESync.checkin(Path,..))
		&& target(source)
		&& args(path,..);
	
	pointcut cancelCheckoutDocument(ESync source, Path path) :
		execution(void ESync.cancelCheckout(Path))
		&& target(source)
		&& args(path);
	

	
	/*
	 * Check when a path is removed from the folders map
	 */
	Boolean around(OutboundProcessor source, Path path, Boolean syncing) : removeFolder(source, path, syncing) {
		Object ret = proceed(source, path, syncing);
		if(ret != null) {
			// This is a new path
			createEvent(source.getESync(), path, 0, "OUTBOUND", "FOLDER-REMOVED", "NONE");
		} 
		return (Boolean)ret;
	}
	
	/*
	 * Check when a new path is added to the folders map if it's new or the syncing value was changed
	 */
	Boolean around(OutboundProcessor source, Path path, Boolean syncing) : addFolder(source, path, syncing) {
		
		Object ret = proceed(source, path, syncing);
		if(ret == null) {
			// This is a new path
			createEvent(source.getESync(), path, 0, "OUTBOUND", "FOLDER-ADDED", "NONE");
			createEvent(source.getESync(), path, 0, "OUTBOUND", "FOLDER-SYNCING-"  + Boolean.toString(syncing).toUpperCase(), "NONE");

		} else {
			// Check if the syncing value changed
			if((Boolean)ret != syncing) {
				createEvent(source.getESync(), path, 0, "OUTBOUND", "FOLDER-SYNCING-" + Boolean.toString(syncing).toUpperCase(), "NONE");
			}
		}
		return (Boolean)ret;
	}
	
	/*
	 * Cancel Checkout Document event
	 */
	before(ESync source, Path path) : cancelCheckoutDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKOUT-CANCEL", "STARTED");
	}
		
	after(ESync source, Path path) : cancelCheckoutDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKOUT-CANCEL", "FINISHED");
	}
	
	/*
	 * Checkin Document event
	 */
	before(ESync source, Path path) : checkinDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKIN", "STARTED");
	}
		
	after(ESync source, Path path) : checkinDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKIN", "FINISHED");
	}
	
	/*
	 * Checkout Document event
	 */
	before(ESync source, Path path) : checkoutDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKOUT", "STARTED");
	}
		
	after(ESync source, Path path) : checkoutDocument(source, path) {
		Path docPath = source.getRepositorySyncPath().resolve(path);
		createEvent(source, docPath, 0, "OUTBOUND", "DOCUMENT-CHECKOUT", "FINISHED");
	}
	
	/*
	 * Delete Document events
	 */
	before(ESync source, Path path) : deleteDocumentInbound(source, path) {
		createEvent(source, source.getFileSystemSyncPath().resolve(path), 0, "INBOUND", "DOCUMENT-DELETE", "STARTED");
	}
	
	after(ESync source, Path path) : deleteDocumentInbound(source, path) {
		createEvent(source, source.getFileSystemSyncPath().resolve(path), 0, "INBOUND", "DOCUMENT-DELETE", "FINISHED");
	}
		
	before(ESync source, Path path) : deleteDocumentOutbound(source, path) {
		createEvent(source, path, 0, "OUTBOUND", "DOCUMENT-DELETE", "STARTED");
	}
		
	after(ESync source, Path path) : deleteDocumentOutbound(source, path) {
		createEvent(source, path, 0, "OUTBOUND", "DOCUMENT-DELETE", "FINISHED");
	}
			
	
	/*
	 * Update Document events
	 */
	before(ESync source, Document document) : updateDocumentInbound(source, document) {
		createEvent(source, source.getRepositoryPathAsFileSystemPath(document), 0, "INBOUND", "DOCUMENT-UPDATE", "STARTED");
	}
	
	after(ESync source, Document document) : updateDocumentInbound(source, document) {
		createEvent(source, source.getRepositoryPathAsFileSystemPath(document), document.getContentStreamLength(), "INBOUND", "DOCUMENT-UPDATET", "FINISHED");
	}
		
	before(ESync source, Path path) : updateDocumentOutbound(source, path) {
		createEvent(source, path, 0, "OUTBOUND", "DOCUMENT-UPDATE", "STARTED");
	}
		
	after(ESync source, Path path) : updateDocumentOutbound(source, path) {
		createEvent(source, path, path.toFile().length(), "OUTBOUND", "DOCUMENT-UPDATE", "FINISHED");
	}
		
	/*
	 * Create Document events
	 */
	before(ESync source, Document document) : createDocumentInbound(source, document) {
		createEvent(source, source.getRepositoryPathAsFileSystemPath(document), 0, "INBOUND", "DOCUMENT-CREATE", "STARTED");
	}
	
	after(ESync source, Document document) : createDocumentInbound(source, document) {
		createEvent(source, source.getRepositoryPathAsFileSystemPath(document), document.getContentStreamLength(), "INBOUND", "DOCUMENT-CREATE", "FINISHED");
	}
	
	before(ESync source, Path path) : createDocumentOutbound(source, path) {
		createEvent(source, path, 0, "OUTBOUND", "DOCUMENT-CREATE", "STARTED");
	}
	
	after(ESync source, Path path) : createDocumentOutbound(source, path) {
		createEvent(source, path, path.toFile().length(), "OUTBOUND", "DOCUMENT-CREATE", "FINISHED");
	}
	
	/*
	 * Sync events
	 */
	
	Object around(InboundProcessor source) : syncInbound(source) {
		createEvent(source.getESync(), source.getSyncPath(), 0, "INBOUND", "SYNC", "STARTED");
		Object ret = proceed(source);
		createEvent(source.getESync(), source.getSyncPath(), 0, "INBOUND", "SYNC", "FINISHED");
		return ret;
	}
	
	Object around(OutboundProcessor source) : syncOutbound(source) {
		createEvent(source.getESync(), source.getSyncPath(), 0, "OUTBOUND", "SYNC", "STARTED");
		Object ret = proceed(source);
		createEvent(source.getESync(), source.getSyncPath(), 0, "OUTBOUND", "SYNC", "FINISHED");
		return ret;
	}
	/*
	 * Create the event and send to the EDrive instance
	 */
	private void createEvent(ESync source, Path path, long size, String direction, String action, String stage) {
		EDriveEvent event = new EDriveEvent(
				source,		// Event source
				action,		// action
				direction,	// direction
				stage,		// stage
				path,		// location
				size,		// size of the File or Document else 0;
				null		// exception
			);
		
		EDrive.getEDrive().processEDriveEvent(event);
	}
}
