/**
 * 
 *    Copyright 2013 EOSSOnline Limited
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  
**/
package com.eossonline.edrive.service;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.regex.Pattern;

import javax.naming.InvalidNameException;

import org.apache.chemistry.opencmis.client.api.CmisObject;
import org.apache.chemistry.opencmis.client.api.Document;
import org.apache.chemistry.opencmis.client.api.DocumentType;
import org.apache.chemistry.opencmis.client.api.Folder;
import org.apache.chemistry.opencmis.client.api.ItemIterable;
import org.apache.chemistry.opencmis.client.api.OperationContext;
import org.apache.chemistry.opencmis.client.api.Rendition;
import org.apache.chemistry.opencmis.client.api.Repository;
import org.apache.chemistry.opencmis.client.api.Session;
import org.apache.chemistry.opencmis.commons.BasicPermissions;
import org.apache.chemistry.opencmis.commons.data.Ace;
import org.apache.chemistry.opencmis.commons.data.Acl;
import org.apache.chemistry.opencmis.commons.data.ContentStream;
import org.apache.chemistry.opencmis.commons.enums.IncludeRelationships;
import org.apache.chemistry.opencmis.commons.exceptions.CmisPermissionDeniedException;
import org.apache.chemistry.opencmis.commons.impl.json.JSONArray;
import org.apache.chemistry.opencmis.commons.impl.json.JSONObject;
import org.apache.chemistry.opencmis.commons.impl.json.parser.JSONParseException;
import org.apache.chemistry.opencmis.commons.impl.json.parser.JSONParser;
import org.apache.commons.codec.binary.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.eossonline.edrive.events.EDriveEvent;
import com.eossonline.edrive.exceptions.EDriveException;
import com.eossonline.edrive.exceptions.UnknownESyncException;
import com.eossonline.esync.ESync;
import com.eossonline.esync.SyncConfiguration;
import com.eossonline.esync.SyncDirectionEnum;
import com.eossonline.esync.exceptions.ShutdownException;
import com.eossonline.esync.exceptions.StartupException;
import com.eossonline.esync.repository.InboundProcessor;

public class EDrive {
	
	private static final Logger logger = LoggerFactory.getLogger(EDrive.class);
	
	private static EDrive eDrive = new EDrive();
	private EDriveMulticastServer ms = null;
	private EDriveCommandServer cs = null;
	
	private String passPhrase = null;

	private ResourceBundle rb = null;

	private Map<String, ESync> synchronisers = new HashMap<String, ESync>(); 
	
	private final List<JSONObject> retryQueue = new ArrayList<JSONObject>();
	private Thread retryQueueThread = null;
	
	private int retryQueueSeconds = 60 * 5; 	// defaults to 5 mins
	
	private EDrive() {
		// Register shutdown handler
		Runtime.getRuntime().addShutdownHook(new Thread() {
		    public void run() {
		        eDrive.close();
		    }
		});
		
		retryQueueThread = new Thread(new Runnable() {
			public void run() {
    			while(!Thread.currentThread().isInterrupted()) {
    				try {
    					Thread.sleep(1000 * retryQueueSeconds);	//  default 5 minutes
    				} catch (InterruptedException ie) {
    	            	Thread.currentThread().interrupt();
    	            	continue;
    				}
    				// Create a temporary queue as we will be removing elements from the original
    				List<JSONObject>tmpQueue = new ArrayList<JSONObject>(retryQueue); 
    				for(JSONObject sync : tmpQueue) {
						if(logger.isDebugEnabled()) {
							logger.debug("Trying to restart Synchroniser [{}].", (String)sync.get("name"));
						}
    					try {
    		    			addSynchroniser(sync);
    		    			retryQueue.remove(sync);
    					}
    		    		catch(Exception se) {
    	    				String serFileName = ESync.getSerFileName(
    	    						Paths.get((String)sync.get("fileSystemRoot")), 
    	    						Paths.get((String)sync.get("repositoryRoot")), 
    	    						(String)sync.get("name"));
    	    				if(ESync.testIfNewSynchroniser((String)sync.get("id"), serFileName, (String)sync.get("name"))) {
    		    				// candidate to add to the retry queue as it has probably never run before and there are probably typos
    		    				logger.error("This is a newly defined Synchroniser [{}] that probably has typos in the URL, Paths or userName password properties. Please recheck all the values you entered.", (String)sync.get("name"));
    		    				// Remove from queue and we won't try this Synchroniser again
    		    				retryQueue.remove(sync);
    		    			} else {
    		    				logger.error(se.getLocalizedMessage() + " EDrive will try to re-start Synchroniser [{}] in {} seconds.", (String)sync.get("name"), retryQueueSeconds);
    		    			}
    		    		}
    				} 
    			}
			}
		});
		
		// Load the resource bundle for the locale
		Locale locale = Locale.getDefault();
		rb = ResourceBundle.getBundle("EDrive", locale);
	}
	
	public String getResourceString(String key) {
		return rb.getString(key);
	}
	
	private void addSynchroniser(ESync es) throws InvalidNameException {
		if(synchronisers.containsKey(es.getName())) {
			throw new InvalidNameException("A Synchroniser with this name [" + es.getName() + "] already exists.");
		}
		for(ESync s : synchronisers.values()) {
			Path fsPath = s.getFileSystemSyncPath();
			Path rPath = s.getRepositorySyncPath();
			
			if(es.getFileSystemSyncPath().startsWith(fsPath) || fsPath.startsWith(es.getFileSystemSyncPath())) {
				String msg = "Invalid Synchroniser configuration. The FileSystem Path for Synchroniser [" 
						+ es.getSyncConfiguration().getName() + "] overlaps with the FileSystem Path for existing Synchroniser [" 
						+ s.getSyncConfiguration().getName() + "].";
				throw new EDriveException(msg);
			} else if(es.getFileSystemSyncPath().startsWith(rPath) || rPath.startsWith(es.getFileSystemSyncPath())) {
				String msg = "Invalid Synchroniser configuration. The Repository Path for Synchroniser [" 
						+ es.getName() + "] overlaps with the Repository Path for existing Synchroniser [" 
						+ s.getName() + "].";
				throw new EDriveException(msg);
			}
		}
		try {
			startup(es);
			synchronisers.put(es.getName(), es);
		} catch(ShutdownException se) {
			logger.error("{}", se.getLocalizedMessage());
		}
	}
	
	public static EDrive getEDrive() {
		return eDrive;
	}
	
	public Collection<ESync> getSynchronisers() {
		return synchronisers.values();
	}
	
	public ESync getSynchroniser(String name) {
		return synchronisers.get(name);
	}
	
	private void removeSynchroniser(ESync es) {
		shutdown(es);
		synchronisers.remove(es.getName());
	}	
	
	public void removeSynchroniser(String name) {
		ESync es = synchronisers.get(name);
		if(es == null) {
	    	throw new UnknownESyncException(name);
		}
		removeSynchroniser(es);
	}
	
	public void shutdown() {
		System.exit(0);
	}
	
	private void close() {
    	logger.warn("Terminating...");

		// TODO : Save the configuration
		// saveSyncConfig();
		
		for(ESync s : synchronisers.values()) {
			shutdown(s);
		}
		
		// Shutdown the Command Server
		cs.shutdown();
		// Shutdown the Multicast Server
		ms.shutdown();

		// Stop the retryQueue thread
		retryQueueThread.interrupt();
		
		logger.warn("Terminated...");
	}
	
	private void shutdown(ESync es) {
		es.shutdown();
	}
	
	private void startup(ESync es) {
		es.startup();
	}

	
	public long getInboundSyncSeconds(String name) {
		ESync es = synchronisers.get(name);
		if(es == null) {
	    	throw new UnknownESyncException(name);
		}
    	return es.getSyncConfiguration().getInboundSyncSeconds();
	}
	
	public long getOutboundSyncSeconds(String name) {
		ESync es = synchronisers.get(name);
		if(es == null) {
	    	throw new UnknownESyncException(name);
		}
		return es.getSyncConfiguration().getOutboundSyncSeconds();
	}
	
	public SyncDirectionEnum getSyncDirection(String name) {
		ESync es = synchronisers.get(name);
		if(es == null) {
	    	throw new UnknownESyncException(name);
		}
		return es.getSyncConfiguration().getSyncDirection();
	}
	
	public void saveSyncConfig() {
		try {
			String defaultLocation = System.getProperty("user.home") + File.separator + "eDrive.json";
			FileWriter file = new FileWriter(System.getProperty("edrive.config.file", defaultLocation));
			file.write(getConfiguration().toJSONString());
			file.flush();
			file.close();
	 
		} catch (IOException e) {
    		logger.error(e.getLocalizedMessage(), e);
		}
	}
	
	private String getPassphraseFromFile(String loc) {
		String passPhrase = null;
		Path ppf = Paths.get(loc);
		try {
    		if(Files.exists(ppf)) {
    			 if(Files.size(ppf) < 8) {
    				 throw new IOException("Pass phase must be 8 or more characters");
    			 }
    			 byte[] passBytes = Files.readAllBytes(ppf);
    			 passPhrase = new String(passBytes);
    		}
		}catch(IOException ioe) {
			logger.error("Error processing pass phrase file [{}].", ppf, ioe);
		}
		return passPhrase;
	}
	
	private void loadSyncConfig(String configFile) throws FileNotFoundException, InvalidNameException {
		
    	JSONParser parser = new JSONParser();
    	 
    	try {
     
    		Object config = parser.parse(new FileReader(configFile));
    		JSONObject jsonObject = (JSONObject) config;
     
    		// loop array
    		JSONArray syncs = (JSONArray) jsonObject.get("edrive");
    		for(Object obj : syncs) {    			
    			JSONObject sync = (JSONObject)obj;    			
    			try {
    				addSynchroniser(sync);
    			}catch(StartupException se) {
    				logger.error(se.getLocalizedMessage() + " If your connection is correct but the Repository is currently unavailable, EDrive will try to re-start Synchroniser [{}] in {} seconds.", (String)sync.get("name"), retryQueueSeconds);
    				addESyncToRetryQueue(sync);    				
    			}
    		}
    		
    	} catch (FileNotFoundException e) {
    		logger.error(e.getLocalizedMessage(), e);
    	} catch (IOException e) {
    		logger.error(e.getLocalizedMessage(), e);
    	} catch (JSONParseException e) {
    		logger.error(e.getLocalizedMessage(), e);
    	}    	
	}
	
	private void addESyncToRetryQueue(JSONObject sync) {
		this.retryQueue.add(sync);
	}
	
	private String getPassPhrase() {
		if(passPhrase == null) {
			
			String defaultPassPhaseLocation = System.getProperty("user.home") + File.separator + "eDrive.pass";
	    	String passPhraseFile = System.getProperty("edrive.passphrase.file", defaultPassPhaseLocation);
	    	if(passPhraseFile != null ) {
	    		passPhrase = getPassphraseFromFile(passPhraseFile);
	    		if(passPhrase == null) {
	            	logger.error("You defined [edrive.passphrase.file] but the file [{}] does not exist.", passPhraseFile);
	    		}
	    	} 
			if(passPhrase == null) { 
				// Set the default pass phrase
	    		logger.warn("Using default passPhrase as one could not be determined.");
	    		passPhrase = InboundProcessor.class.toString();
	    	}
			
		} 
		return passPhrase;
	}
	
	public ESync addSynchroniser(JSONObject sync) {
		ESync eSync = new ESync(new SyncConfiguration(sync), getPassPhrase());
		
		// Disable synching of any Paths defined as not synching
		JSONArray ja = (JSONArray)sync.get("paths");
		for(Object pobj : ja) {
			JSONObject jo = (JSONObject)pobj;
			eSync.setSyncPath(Paths.get((String)jo.get("path")), (Boolean)jo.get("syncing"));
		}
		try {
			addSynchroniser(eSync);
		} catch(Exception e) {
			throw new EDriveException(e);
		}
		return eSync;
	}

    public JSONObject getConfiguration() {
    	JSONObject config = new JSONObject();

    	JSONArray ja = new JSONArray();
		for(ESync es : synchronisers.values()) {
			ja.add(es.getJSONSyncConfiguration());
		}
    	config.put("esync", ja);

    	return config;
    }
    
    public JSONArray getAvailableFolders(Repository repo, Path startFrom) {
    	JSONArray availableFolders = new JSONArray();
    	
    	try {
	    	Session repoSession = repo.createSession();

	    	CmisObject folder = null;
	    	if(startFrom != null) {
	    		try {
	    			folder = repoSession.getObjectByPath(ESync.getPathAsString(startFrom));
	    			if(!(folder instanceof Folder)) {
	    				throw new Exception();
	    			}
	    		}catch(Exception e) {
	    			folder = null;
	    		}
	    	}
	    	if(folder == null){
	    		// Get from the root folder
	    		folder = repoSession.getRootFolder();
	    	}
	    	availableFolders.add(getAvailableFolders(repoSession, (Folder)folder));
	    } catch(Exception e) {
    		logger.error("Exception while getting Repository Folders", e);
    	}

    	return availableFolders;
    }
    
    private JSONObject getAvailableFolders(Session session, Folder current) {
    	JSONObject folders = new JSONObject();

    	ItemIterable<CmisObject> it = current.getChildren();
    	for(CmisObject obj : it) {
    		if(obj instanceof Folder) {
    			JSONObject tmp = getAvailableFolders(session, (Folder)obj);
    			for(String key : tmp.keySet()) {
    				folders.put(key, (Boolean)tmp.get(key));
    			}
    		}
    	}
    	
		Boolean write = true; // Start off assuming write permission
		// Check if we have write permission
		
		try {
			Acl acl = session.getAcl(current, true);
			if(acl != null) {
				write = false; // Because we have ACL's assume no write
    			List<Ace> aces = acl.getAces();

    			for(Ace ace : aces) {
    				if(ace.getPermissions().contains(BasicPermissions.WRITE)) {
    					write = true;
    					break;
    				}
    			}
			}
		}catch(CmisPermissionDeniedException cpde) {
			write = false; // Assume that we do not have ANY permissions
		}
		folders.put(((Folder)current).getPath(), write);
    	
    	return folders;
    }
    
	public JSONObject getConfigurationDefaults() {
		SyncConfiguration sc = SyncConfiguration.getDefaults();;
		
		JSONObject config = new JSONObject();

		Base64 base64 = new Base64();
		JSONArray ignores = new JSONArray();
		for(Pattern pattern : sc.getIgnorePatterns()) {
			JSONObject ignore = new JSONObject();
			ignore.put("base64", new String(base64.encode(pattern.pattern().getBytes())));
			ignores.add(ignore);
		}
		if(!ignores.isEmpty()) {
			// We may at some point allocate special file patterns we don't want synched here
			config.put("ignorePatterns", ignores);
		}

		config.put("inboundSyncSeconds", sc.getInboundSyncSeconds());
		config.put("outboundSyncSeconds", sc.getOutboundSyncSeconds());
		config.put("deleteEnabled", sc.isRepositoryDeleteEnabled());
		config.put("unfilingEnabled", sc.isUnfilingEnabled());
		config.put("syncDirection", sc.getSyncDirection().toString());
		
		return config;
	}

    /**
     * Process an ESyncEvent
     * @param event
     */
    public void processEDriveEvent(EDriveEvent event) {
    	if(ms != null) {
    		ms.addEvent(event);
    	}
    }
    
    private EDriveMulticastServer addMulticastServer(EDriveMulticastServer ms) {
    	EDriveMulticastServer ret = ms;
    	this.ms = ms;
    	return ret;
    }
    
    private EDriveCommandServer addCommandServer(EDriveCommandServer cs) {
    	EDriveCommandServer ret = cs;
    	this.cs = cs;
    	return ret;
    }
    
    public JSONObject updateConfiguration(JSONObject obj) {
    	JSONArray syncs = new JSONArray();
    	
    	for(ESync eSync : synchronisers.values()) {
    		syncs.add(eSync.updateConfiguration(obj));
    	}
    	JSONObject result = new JSONObject();
    	result.put("eSync", syncs);
    	return result;
    }
    
    public JSONObject setConfiguration(JSONObject obj) {
    	JSONArray syncs = new JSONArray();
    	
    	for(ESync eSync : synchronisers.values()) {
    		syncs.add(eSync.setConfiguration(obj));
    	}
    	JSONObject result = new JSONObject();
    	result.put("eSync", syncs);
    	return result;
    }
    
    /**
     *  Get the Version history of a document.
     *  
     *  Supports either the File system path, the Document Path or Relative path which 
     *  is the same for both Repository and FileSystem locations
     *  
     *  If requesting a Document that is not versionable, return details about the this document
     *  
     *  If requesting details about a Folder, return the folder details
     *  
     *  If requesting details about a non-existent document or folder an exception is thrown
     */
    
    public JSONObject getVersionHistory(ESync eSync, Path path) {
    	JSONObject result = new JSONObject();
    	Path rPath = null;
    	Path fPath = null;
    	
    	if(path.isAbsolute()) {
    		// Work out if it's a file system or repository path
    		if(path.startsWith(eSync.getFileSystemSyncPath())) {
    			// It's a file system path
    			rPath = eSync.getRepositorySyncPath().resolve(eSync.getFileSystemSyncPath().relativize(path));
    			fPath = path;
    		} else if(path.startsWith(eSync.getRepositorySyncPath())){
    			// It's a Repository path
    			rPath = path;
    			fPath = eSync.getFileSystemSyncPath().resolve(eSync.getRepositorySyncPath().relativize(path));
    		} else {
    			// No idea what you are
    			throw new EDriveException("The Path [" + path + "] Cannot be found.");
    		}
    	} else {
    		// it's a relative path so get repository full path
    		rPath = eSync.getRepositorySyncPath().resolve(path);
    		fPath = eSync.getFileSystemSyncPath().resolve(path);
    	}
    	
    	CmisObject obj = eSync.getInboundProcessor().getSession().getObjectByPath(ESync.getPathAsString(rPath));
    	
    	JSONArray ja = new JSONArray();
    	SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
    	
    	if(obj instanceof Document) {
    		Document doc = (Document)obj;
    	
    		if(((DocumentType)doc.getType()).isVersionable()) {
		        List<Document> versions = doc.getAllVersions();
		        for (Document version : versions) {
		        	JSONObject ver = new JSONObject();
		        	ver.put("id", version.getId());
		        	ver.put("name", version.getName());
		        	ver.put("versionLabel", version.getVersionLabel());
		        	ver.put("versionSeriesId", version.getVersionSeriesId());
		        	ver.put("checkedOutBy", version.getVersionSeriesCheckedOutBy());
		        	ver.put("checkedOutId", version.getVersionSeriesCheckedOutId());
		        	ver.put("majorVersion", version.isMajorVersion());
		        	ver.put("latestVersion", version.isLatestVersion());
		        	ver.put("latestMajorVersion", version.isLatestMajorVersion());
		        	ver.put("checkinComment", version.getCheckinComment());
		        	ver.put("contentLength", version.getContentStreamLength());
		        	ver.put("contentStreamFileName", version.getContentStreamFileName());
		        	Calendar cal = (Calendar)version.getLastModificationDate();
		        	ver.put("lastModificationDate",sdf.format(cal.getTime()));
		        	ver.put("lastModifiedBy",version.getLastModifiedBy());
		        	ver.put("mimeType", version.getContentStreamMimeType());
		        	ver.put("type", version.getType().getId());
		        	
		        	ja.add(ver);
		        }
    		} else {
	        	JSONObject ver = new JSONObject();
	        	ver.put("id", doc.getId());
	        	ver.put("name", doc.getName());
	        	ver.put("contentLength", doc.getContentStreamLength());
	        	Calendar cal = (Calendar)doc.getLastModificationDate();
	        	ver.put("lastModificationDate",sdf.format(cal.getTime()));
	        	ver.put("lastModifiedBy",doc.getLastModifiedBy());
	        	ver.put("mimeType", doc.getContentStreamMimeType());
	        	ver.put("type", doc.getType().getId());
	        	
	        	ja.add(ver);
    		}
    	} else if (obj instanceof Folder) {
    		Folder folder = (Folder)obj;
    		JSONObject ver = new JSONObject();
    		ver.put("id", folder.getId());
    		ver.put("type", folder.getType().getId());
    		ver.put("name", folder.getName());
        	Calendar cal = (Calendar)obj.getLastModificationDate();
    		ver.put("lastModificationDate", sdf.format(cal.getTime()));
    		ver.put("lastModifiedBy", folder.getLastModifiedBy());
    		
    		ja.add(ver);
    		
    	} else {
    		// Dont know what this is but we donta handle it
    		throw new EDriveException("The Path [" + path + "] does not point to a Folder or Document.");
    	}

    	result.put("repositoryPath", ESync.getPathAsString(rPath));
    	result.put("fileSystemPath", fPath.toString());
    	result.put("synchroniserName", eSync.getName());
    	result.put("synchroniserId", eSync.getSyncConfiguration().getId());
    	result.put("versions", ja);
    	
    	return result;
    }
    
    public JSONObject getThumbNails(ESync eSync, Path path) {
    	JSONObject result = new JSONObject();

    	Path rPath = null;
    	Path fPath = null;
    	
    	if(path.isAbsolute()) {
    		// Work out if it's a file system or repository path
    		if(path.startsWith(eSync.getFileSystemSyncPath())) {
    			// It's a file system path
    			rPath = eSync.getRepositorySyncPath().resolve(eSync.getFileSystemSyncPath().relativize(path));
    			fPath = path;
    		} else if(path.startsWith(eSync.getRepositorySyncPath())){
    			// It's a Repository path
    			rPath = path;
    			fPath = eSync.getFileSystemSyncPath().resolve(eSync.getRepositorySyncPath().relativize(path));
    		} else {
    			// No idea what you are
    			throw new EDriveException("The Path [" + path + "] Cannot be found.");
    		}
    	} else {
    		// it's a relative path so get repository full path
    		rPath = eSync.getRepositorySyncPath().resolve(path);
    		fPath = eSync.getFileSystemSyncPath().resolve(path);
    	}
    	
    	OperationContext docOpCtx = eSync.getInboundProcessor().getSession().createOperationContext();
    	docOpCtx.setIncludeAcls(false);
    	docOpCtx.setIncludeAllowableActions(true);
    	docOpCtx.setIncludePolicies(false);
    	docOpCtx.setIncludeRelationships(IncludeRelationships.NONE);
    	docOpCtx.setRenditionFilterString("cmis:thumbnail");
    	docOpCtx.setIncludePathSegments(false);
    	docOpCtx.setOrderBy(null);
    	docOpCtx.setCacheEnabled(true);

    	CmisObject obj = eSync.getInboundProcessor().getSession().getObjectByPath(ESync.getPathAsString(rPath), docOpCtx);
    	
    	JSONArray ja = new JSONArray();
    	
    	List<Rendition> renditions = obj.getRenditions();
    	if(renditions == null || renditions.isEmpty()) {
    		throw new EDriveException("There are no Thumb Nails available for [" + ESync.getPathAsString(rPath) + "]");
    	}
    	for(Rendition r : renditions) {
    		JSONObject jo = new JSONObject();
    		
    		jo.put("height", r.getHeight());
    		jo.put("width", r.getWidth());
    		jo.put("title", r.getTitle());
    		jo.put("length", r.getLength());
    		jo.put("mimeType", r.getMimeType());
    		jo.put("id", r.getRenditionDocumentId());
    		
    		ja.add(jo);
    	}
    	result.put("repositoryPath", ESync.getPathAsString(rPath));
    	result.put("fileSystemPath", fPath.toString());
    	result.put("thumbNails", ja);
    	return result;
    }
    
    public JSONObject getDocument(ESync eSync, String documentId) {
    	JSONObject result = new JSONObject();
    	
    	try {
    		CmisObject obj = eSync.getInboundProcessor().getSession().getObject(documentId);
    		if(!(obj instanceof Document)) {
    			throw new EDriveException("The requested id [" + documentId + "] does not point to a Document.");
    		}
    		Document doc = (Document)obj;
    		ContentStream cs = doc.getContentStream();
    		result.put("name",cs.getFileName());
    		result.put("mimeType", cs.getMimeType());
    		result.put("id", documentId);
    		result.put("contentLength", cs.getLength());
    		
    		ByteArrayOutputStream out = new ByteArrayOutputStream();
    		byte[] buffer = new byte[1024]; // you can configure the buffer size
    		int length = 0;

    		InputStream is = cs.getStream();
    		try {
	    		while ((length = is.read(buffer)) != -1) 
	    			out.write(buffer, 0, length); //copy streams
    		}finally {
    			try {
    				is.close();
    			}catch(Exception e) {
    				throw new EDriveException(e);
    			}
    		}
    		// We now have the file as a byte array so encode as base64 and put in result   		
    		result.put("dataBase64", Base64.encodeBase64String(out.toByteArray()));
    		
    	} catch(Exception e) {
    		throw new EDriveException(e);
    	}
    	
    	return result;
    }
    
    public static void main(String[] args) throws Exception {
        
    	int port = Integer.parseInt(System.getProperty("edrive.server.port", "0"));
    	try {
    		eDrive.addCommandServer(new EDriveCommandServer(port));
    	}catch(Exception e) {
    		logger.error("Command Server NOT available...");
    	}
    	
    	// The default is 300 seconds = 5 minutes
    	eDrive.retryQueueSeconds = Integer.parseInt(System.getProperty("edrive.retry.seconds", "300"));
    	
    	// Calculate the multicast address if not specified
    	// String multicatAddress = "224.0.0.1";
    	String multicastAddress = System.getProperty("edrive.multicast.ip");
    	if(multicastAddress == null) {
    		logger.error("You MUST define a Multicast IP Address to enable the Multicast Server. It will not be available.");
    	} else {
	    	int multicastPort = Integer.parseInt(System.getProperty("edrive.multicast.port", "0"));
	    	try {
	    		eDrive.addMulticastServer(new EDriveMulticastServer(eDrive, multicastAddress, multicastPort));
	    	}catch(Exception e) {
	    		logger.error("Multicast Server NOT available...");
	    	}
    	}
    	
        // Load and start any previously saved configurations
    	eDrive.loadSyncConfig(System.getProperty("edrive.config.file", "eDrive.json"));

    	// Start the retry Queue thread
		eDrive.retryQueueThread.run();
    }
}

