/**
 * 
 *    Copyright 2013 EOSSOnline Limited
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  
**/
package com.eossonline.edrive.service;

import java.io.*;
import java.net.*;
import java.util.ArrayList;
import java.util.List;

import org.apache.chemistry.opencmis.commons.impl.json.JSONObject;
import org.apache.chemistry.opencmis.commons.impl.json.parser.JSONParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.eossonline.edrive.commands.CommandFactory;

public class EDriveCommandServer extends Thread {

	private static final Logger logger = LoggerFactory.getLogger(EDriveCommandServer.class);

	private ServerSocket serverSocket;
	private int port = 0;
	
	public EDriveCommandServer(int port) {
		try {
			serverSocket = new ServerSocket(port);
			this.port = serverSocket.getLocalPort();
			// create new ServerSocket
		} catch (IOException e) {
			logger.error("Unable to start the ESyncServer. No external command connection possible...");
			logger.error(e.getLocalizedMessage(), e);
			throw new RuntimeException(e);
		}
		this.start();
	}
	public void run() {
		logger.info("Listening on port [{}] for connections...", this.port);

		List<Socket> clients = new ArrayList<Socket>();
		
		// repeatedly wait for connections, and process
		Socket clientSocket = null;
		while (!Thread.currentThread().isInterrupted()) {
			try {
				// a "blocking" call which waits until a connection is requested
				clientSocket = serverSocket.accept();
				final Socket _clientSocket = clientSocket;
				Thread t = new Thread(new Runnable() {
					public void run() {
						processRequests(_clientSocket);
					}
				});
				clients.add(_clientSocket);
				t.start();

			} catch (IOException ioe) {
				if(serverSocket.isClosed()) {
					break;
				}
				logger.error("Error while waiting for client connection...", ioe);
			} 
		}
		if(!clients.isEmpty()) {
			logger.warn("Terminating client connections...");
			for(Socket s : clients) {
				try {
					s.close();
				}catch(IOException ignore) {}
			}
			logger.warn("Terminated client connections...");
		}
	}
	
	protected void shutdown() {
		logger.warn("Terminating...");
		try {
			serverSocket.close();
		} catch(IOException ioe) {
			logger.error("Exception while closing the ServerSocket.", ioe);
		}
		logger.warn("Terminated...");
	}
	
	private void processRequests(Socket clientSocket) {

		JSONParser parser = new JSONParser();

		logger.info("Accepted connection from client [{}].", clientSocket.getInetAddress().getHostAddress());

		BufferedReader in = null;
		PrintWriter out = null;

		try {
			// open up IO streams
			in = new BufferedReader(new InputStreamReader(
					clientSocket.getInputStream()));
			out = new PrintWriter(new OutputStreamWriter(
					clientSocket.getOutputStream()));

			out.println("OK awaiting commands");
			out.println();
			out.flush();

			// waits for data and reads it in until connection dies
			// readLine() blocks until the server receives a new line
			String s;
			String result;
			while (((s = in.readLine()) != null)) {
				if(s == null || s.trim().length() == 0) {
					continue;
				}
				s = s.trim();
				
				s = unJSONCommand(parser, s);
				
				if(logger.isDebugEnabled()) {
					logger.debug("Client [{}] requested [{}]", clientSocket.getInetAddress().getHostAddress(), s);
				}
				if(s.equalsIgnoreCase("quit")) {
					result = quit();
				} else if(s.equalsIgnoreCase("help")) {
					result = help();
				}else if(s.equalsIgnoreCase("licence")) {
					result = licence();
				}
				else {
					result = CommandFactory.getInstance().processRequest(parser, s).toString();
				}
				out.println(result);
				out.flush();
				if(result.startsWith("Good by")) {
					break;
				}
			}
		} catch(SocketException se) {
			logger.info("Socket for client connection [[]] has been forcefully closed...", clientSocket.getInetAddress().getHostAddress());
		} catch (IOException ioe) {
			logger.error("Error while processing commands for client [{}]. Shutting down connection..."
					,clientSocket.getInetAddress().getHostAddress(), ioe);
			out.println(CommandFactory.generateExceptionResponse(ioe));
			out.println("Good by...");
			out.flush();
		} finally {
			if(!clientSocket.isClosed()) {
				try {
					out.close();
					in.close();
					clientSocket.close();
				} catch (IOException ioe) {
					logger.error(ioe.getLocalizedMessage(), ioe);
				}
			}
		}
		logger.info("Connection with client [{}] closed ...", clientSocket.getInetAddress().getHostAddress());
	}

	private String unJSONCommand(JSONParser parser, String s) {
		
		try {
			JSONObject obj = (JSONObject)parser.parse(s);
			String command = (String)obj.get("command");
			JSONObject params = (JSONObject)obj.get("params");
			
			s = command + (params == null ? "" : " " + params.toString());
			
		}catch(Exception ignore) {
			// It wasn't a JSON object version of the command
		}
		
		return s;
	}
	
	private String licence() {
		String licenceFileName = EDrive.getEDrive().getResourceString("edrive.licence.file");
		BufferedReader br = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResourceAsStream(licenceFileName)));

		StringBuilder sb = new StringBuilder();

		String line = null;
		try {
			while((line = br.readLine()) != null) {
				sb.append(line).append("\n\r");
			}
		}catch(IOException ioe) {
			return "Problem reading licence file [" + licenceFileName + "] from classpath.  Make sure the edrive.licence.file entry in the EDrive.properties file is correct.";
		} finally {
			try {
				br.close();
			} catch(Exception ignore) {}
		}

		return sb.toString();
	}
	
	/*
	 * Get the help text from the help.txt file on the classpath
	 */
	private String help() {
		String helpFileName = EDrive.getEDrive().getResourceString("edrive.help.file");
		BufferedReader br = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResourceAsStream(helpFileName)));

		StringBuilder sb = new StringBuilder();

		String line = null;
		try {
			while((line = br.readLine()) != null) {
				sb.append(line).append("\n\r");
			}
		}catch(IOException ioe) {
			return "Problem reading help file [" + helpFileName + "] from classpath. Make sure the edrive.help.file entry in the EDrive.properties file is correct.";
		} finally {
			try {
				br.close();
			} catch(Exception ignore) {}
		}

		return sb.toString();
		
		/*
		sb.append("All commands are case-insensitive").append("\n\r\n\r");
		sb.append("\t").append("To get help type Help.").append("\n\r");
		sb.append("\t").append("To quit type Quit");
		return sb.toString();
		*/
	}

	private String quit() {
		return "Good by...";
	}
	

}
